import os
import logging
import requests
from bs4 import BeautifulSoup
from langchain_core.tools import tool
from googleapiclient.discovery import build
from .db_core import get_app_settings, get_domain
from .db_agent_ops import is_url_processed, record_scrape_failure, is_domain_system_blacklisted
from .db_stats import log_api_call
import feedparser
from urllib.parse import urljoin

logger = logging.getLogger(__name__)

# --- Database Check Tool ---

@tool
def check_if_url_is_processed(url: str) -> bool:
    """
    Checks if a given URL has already been processed and stored previously.
    Returns True if the URL has been processed, False otherwise.
    Use this tool BEFORE processing any article found via search to avoid redundant work.
    """
    return is_url_processed(url)


# --- Google Search Tool ---

@tool
def google_search(query: str, num_results: int = 10, date_restrict: str = 'd1') -> list[dict]:
    """
    Performs a Google Custom Search for the given query and returns a list of results.
    Each result is a dictionary containing 'title', 'link', and 'snippet'.
    Requires GOOGLE_API_KEY and GOOGLE_CSE_ID to be configured in the application settings.

    Args:
        query: The search query string.
        num_results: The total number of results desired. The tool will automatically handle
                     making multiple API calls in batches of 10 if num_results > 10.
        date_restrict: Restricts search results to those published within the specified time period
                       (e.g., 'd1' for last 24 hours, 'w1' for last week, 'm1' for last month).
    """
    logger.debug(f"Tool: Performing Google Search for query: '{query}' (max {num_results} results)")
    google_api_key = get_app_settings().get('google')
    google_cse_id = get_app_settings().get('google_cse_id')

    if not google_api_key or not google_cse_id:
        logger.error("Google API Key or CSE ID is missing. Cannot perform search.")
        log_api_call(
            api_provider='Google',
            service_model='Custom Search',
            success=False,
            caller_component='Collector',
            metadata={'query': query, 'error': 'Search tool is not configured. Missing Google API Key or CSE ID.'}
        )
        return [{"error": "Search tool is not configured. Missing Google API Key or CSE ID."}]

    all_results = []
    max_per_request = 10 # Google CSE API limit per request
    start_index = 1
    success = False # Track overall success for logging

    try:
        service = build("customsearch", "v1", developerKey=google_api_key)

        while start_index <= num_results:
            # Calculate how many results to request in this specific API call
            num_to_fetch = min(max_per_request, num_results - start_index + 1)
            if num_to_fetch <= 0:
                break # Should not happen with the loop condition, but safe check

            res = service.cse().list(
                q=query,
                cx=google_cse_id,
                num=num_to_fetch,
                start=start_index,
                dateRestrict=date_restrict
            ).execute()

            items = res.get('items', [])
            batch_results = [{"title": item.get('title'), "link": item.get('link'), "snippet": item.get('snippet')} for item in items]
            all_results.extend(batch_results)

            # Check if we received fewer results than requested, indicating no more results are available
            if len(items) < num_to_fetch or 'nextPage' not in res.get('queries', {}):
                break

            start_index += max_per_request # Move to the next batch start index

        success = True # Mark as successful if loop completes without error

    except Exception as e:
        # Log the error but return any results gathered so far, plus an error indicator
        logger.error(f"Google Search failed for query '{query}': {e}", exc_info=True)
        all_results.append({"error": f"An error occurred during Google Search after fetching {len(all_results)} results: {e}"})
        success = False # Mark as failed
        # Log the failure immediately
        log_api_call(
            api_provider='Google',
            service_model='Custom Search',
            success=False,
            caller_component='Collector', # Assume Collector calls this
            metadata={'query': query, 'error': str(e)}
        )

    # Log success if it wasn't logged as failure during exception
    if success:
        log_api_call(api_provider='Google', service_model='Custom Search', success=True, caller_component='Collector', metadata={'query': query, 'results_count': len(all_results)})

    return all_results


# --- Web Scraping Tool ---

@tool
def scrape_web_content(url: str) -> str:
    """
    Fetches the content of a given URL and extracts the main text content using BeautifulSoup.
    Returns the extracted text or an error message if fetching/parsing fails.
    """
    domain = get_domain(url) # Get domain for potential error logging
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'}
    try:
        response = requests.get(url, headers=headers, timeout=10)
        response.raise_for_status() # Raise an exception for bad status codes (4xx or 5xx)

        # Use BeautifulSoup to parse the HTML content
        soup = BeautifulSoup(response.text, 'html.parser')

        # Attempt to find the main content - this is heuristic and might need refinement
        # Look for common tags like <article>, <main>, or divs with relevant IDs/classes
        main_content = soup.find('article') or soup.find('main') or soup.find('div', {'role': 'main'})
        if not main_content:
             # Fallback: try to get text from the body, removing script/style tags
             for tag in soup(['script', 'style', 'nav', 'footer', 'header', 'aside']):
                 tag.decompose()
             main_content = soup.body

        text = main_content.get_text(separator='\n', strip=True) if main_content else "Could not extract main content."
        return text

    except requests.exceptions.RequestException as e:
        error_type = f"{e.__class__.__name__}" # e.g., "ConnectionError"
        if hasattr(e, 'response') and e.response is not None:
            error_type = f"{e.response.status_code} {e.response.reason}" # e.g., "403 Forbidden"
        # Record the failure
        if domain: record_scrape_failure(url, domain, error_type)
        return f"Error: Could not fetch URL. {e}"
    except Exception as e:
        error_type = f"Scrape Error: {e.__class__.__name__}"
        # Record the failure
        if domain: record_scrape_failure(url, domain, error_type)
        return f"Error: Could not parse content from URL. {e}"

# --- RSS Feed Tools ---

@tool
def find_rss_feed(site_url: str) -> str | None:
    """
    Tries to automatically discover the RSS or Atom feed URL for a given website URL.
    Checks common locations and <link> tags in the HTML head.
    Returns the feed URL if found, otherwise None.
    """
    headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'}
    try:
        response = requests.get(site_url, headers=headers, timeout=10)
        response.raise_for_status()
        soup = BeautifulSoup(response.text, 'html.parser')

        # Look for <link> tags with type="application/rss+xml" or "application/atom+xml"
        feed_links = soup.find_all('link', type=['application/rss+xml', 'application/atom+xml'])
        if feed_links:
            feed_url = feed_links[0].get('href')
            if feed_url:
                absolute_feed_url = urljoin(site_url, feed_url) # Ensure URL is absolute
                return absolute_feed_url

        # Add checks for common feed paths if needed (e.g., /feed, /rss.xml) - less reliable
        # common_paths = ['/feed', '/rss', '/atom.xml', '/feed.xml', '/rss.xml']
        # for path in common_paths:
        #     potential_url = urljoin(site_url, path)
        #     try:
        #         # Quick HEAD request to check existence
        #         head_response = requests.head(potential_url, headers=headers, timeout=5, allow_redirects=True)
        #         if head_response.status_code == 200 and 'xml' in head_response.headers.get('Content-Type', ''):
        #              return potential_url
        #     except requests.exceptions.RequestException:
        #         continue # Ignore errors for common path checks

        return None

    except requests.exceptions.RequestException as e:
        logger.error(f"Failed to fetch site {site_url} to find RSS feed: {e}")
        return None
    except Exception as e:
        logger.error(f"Error parsing HTML for RSS feed at {site_url}: {e}", exc_info=True)
        return None

@tool
def parse_rss_feed(feed_url: str) -> list[dict]:
    """
    Parses an RSS or Atom feed URL and extracts news items.
    Returns a list of dictionaries, each containing 'title', 'link', 'summary', and 'published'.
    Returns an empty list if parsing fails or the feed is empty.
    """
    try:
        # feedparser handles fetching and parsing
        feed_data = feedparser.parse(feed_url, agent='Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36')
        if feed_data.bozo: # Check for errors during parsing
            logger.warning(f"Feedparser encountered issues parsing {feed_url}: {feed_data.bozo_exception}")

        entries = []
        for entry in feed_data.entries:
            entries.append({
                "title": entry.get("title", "No Title"),
                "link": entry.get("link"), # This is the canonical article URL
                "summary": entry.get("summary", entry.get("description", "No Summary")), # Use description as fallback
                "published": entry.get("published_parsed", entry.get("updated_parsed")) # Get parsed time tuple
            })
        return entries
    except Exception as e:
        logger.error(f"Failed to parse RSS feed {feed_url}: {e}", exc_info=True)
        return []

# List of tools commonly used by agents
common_tools = [check_if_url_is_processed]
collector_tools = [google_search, check_if_url_is_processed, scrape_web_content, find_rss_feed, parse_rss_feed]
